/**
 * Default implementation of {@link TokenStorage}, delegating to Storage of the browser.
 * (`sessionStorage` or `localStorage`).
 *
 * It should be given to {@link Auth} constructor through {@link AuthOptions.tokenStorage} and
 * {@link AuthOptions.persistentTokenStorage}.
 */
var DefaultTokenStorage = /** @class */ (function () {
    function DefaultTokenStorage(storage, accessTokenStorageKey, refreshTokenStorageKey, expiresAtSuffix, credentialsTokenStorageKey) {
        if (accessTokenStorageKey === void 0) { accessTokenStorageKey = 'auth.accessToken'; }
        if (refreshTokenStorageKey === void 0) { refreshTokenStorageKey = 'auth.refreshToken'; }
        if (expiresAtSuffix === void 0) { expiresAtSuffix = '.expiresAt'; }
        if (credentialsTokenStorageKey === void 0) { credentialsTokenStorageKey = 'auth.credentials'; }
        this.async = false;
        this.storage = storage;
        this.accessTokenStorageKey = accessTokenStorageKey;
        this.refreshTokenStorageKey = refreshTokenStorageKey;
        this.expiresAtSuffix = expiresAtSuffix;
        this.credentialsTokenStorageKey = credentialsTokenStorageKey;
    }
    DefaultTokenStorage.prototype.store = function (tokens) {
        this.storage.setItem(this.accessTokenStorageKey, tokens.access.value);
        if (tokens.access.expiresAt) {
            this.storage.setItem(this.accessTokenStorageKey + this.expiresAtSuffix, tokens.access.expiresAt.getTime().toString(10));
        }
        else {
            this.storage.removeItem(this.accessTokenStorageKey + this.expiresAtSuffix);
        }
        if (tokens.refresh) {
            this.storage.setItem(this.refreshTokenStorageKey, tokens.refresh.value);
        }
        else {
            this.storage.removeItem(this.refreshTokenStorageKey);
        }
        if (tokens.refresh && tokens.refresh.expiresAt) {
            this.storage.setItem(this.refreshTokenStorageKey + this.expiresAtSuffix, tokens.refresh.expiresAt.getTime().toString(10));
        }
        else {
            this.storage.removeItem(this.refreshTokenStorageKey + this.expiresAtSuffix);
        }
        if (tokens.credentials) {
            this.storage.setItem(this.credentialsTokenStorageKey, JSON.stringify(tokens.credentials));
        }
        else {
            this.storage.removeItem(this.credentialsTokenStorageKey);
        }
    };
    DefaultTokenStorage.prototype.clear = function () {
        this.storage.removeItem(this.accessTokenStorageKey);
        this.storage.removeItem(this.refreshTokenStorageKey);
        this.storage.removeItem(this.accessTokenStorageKey + this.expiresAtSuffix);
        this.storage.removeItem(this.refreshTokenStorageKey + this.expiresAtSuffix);
        this.storage.removeItem(this.credentialsTokenStorageKey);
    };
    DefaultTokenStorage.prototype.getTokens = function () {
        var accessTokenStr = this.storage.getItem(this.accessTokenStorageKey);
        var refreshTokenStr = this.storage.getItem(this.refreshTokenStorageKey);
        var accessTokenExpiresAtStr = this.storage.getItem(this.accessTokenStorageKey + this.expiresAtSuffix);
        var refreshTokenExpiresAtStr = this.storage.getItem(this.refreshTokenStorageKey + this.expiresAtSuffix);
        var credentialsStr = this.storage.getItem(this.credentialsTokenStorageKey);
        var accessTokenExpiresAt = accessTokenExpiresAtStr ? new Date(parseInt(accessTokenExpiresAtStr, 10)) : undefined;
        var refreshTokenExpiresAt = refreshTokenExpiresAtStr
            ? new Date(parseInt(refreshTokenExpiresAtStr, 10))
            : undefined;
        var accessToken = accessTokenStr ? accessTokenStr : undefined;
        var refreshToken = refreshTokenStr ? refreshTokenStr : undefined;
        var credentials = credentialsStr ? JSON.parse(credentialsStr) : undefined;
        if (accessToken) {
            var tokens = { access: { value: accessToken, expiresAt: accessTokenExpiresAt } };
            if (refreshToken) {
                tokens.refresh = { value: refreshToken, expiresAt: refreshTokenExpiresAt };
            }
            if (credentialsStr) {
                tokens.credentials = credentials;
            }
            return tokens;
        }
    };
    return DefaultTokenStorage;
}());
export { DefaultTokenStorage };
//# sourceMappingURL=default-token-storage.js.map